/*
 * cache.c
 *
 * Copyright (C) 1994, John Kilburg.
 *
 * You may distribute this code freely as long as it is used only
 * for academic, research, and internal business uses only, without
 * a fee.  You may distribute the binary and source code to third
 * parties provided that the copyright notice and this statement
 * appears on all copies and that no charge is associated with such
 * copies.
 *
 * If you make and distribute modified versions of the source or
 * binaries then you must inform me (John Kilburg) of the distribution.
 * You must also give me credit for the original in the source and
 * documentation but you must indicate that modifications were made.
 *
 * If you wish to make commercial use of the source or binaries you should    
 * contact me, to negotiate an appropriate license for such
 * commercial use.  Commercial use includes (1) integration of all or    
 * part of the source code into a product for sale or license by or on   
 * behalf of Licensee to third parties, or (2) distribution of the binary
 * code or source code to third parties that need it to utilize a          
 * commercial product sold or licensed by you or on your behalf.        
 * 
 * This software is provided "as is" without expressed or
 * implied warranty.  It is not intended to be used for any particular
 * purpose.  I shall not be liable for any damages suffered by users
 * of this software.  USE AT YOUR OWN RISK.
 *
 */
#include <stdio.h>
#include <ctype.h>

#ifndef NOSTDHDRS
#include <stdlib.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/param.h>

#include <fcntl.h>

#if defined(SYSV) || defined(SVR4)
#include <dirent.h>
#else
#include <sys/dir.h>
#endif

/* Jim Rees fix */
#ifndef S_ISDIR
#define S_ISDIR(mode) (((mode) & S_IFMT) == S_IFDIR)
#endif

#include "url.h"
#include "document.h"
#include "util.h"
#include "conf.h"

struct cache_info
{
  char *path;
  time_t atime;
  off_t size;
  struct cache_info *next;
};

static unsigned long cache_size = 0;
static char *cache_dir = "/tmp";
static struct cache_info *clist = NULL;
static time_t doc_ttl = 0;
static unsigned long max_cache_size = 0;

/*
 * cache information
 *
 * the document cache is a disk-based cache with an in-core list
 * to keep track of its contents.  since the cache survives between
 * sessions InitCache() must be used to build the index.  the index
 * is used when the cache is turned off to keep track of which links
 * have been visited.
 *
 * if a document is older than doc_ttl then it is reloaded from the
 * net.  if the cache grows beyond max_cache_size then the cache file
 * that was accessed least recently is removed.
 *
 * if doc_ttl is zero then documents are not flushed because of age.
 */

/*
 * InsertCacheInfo
 */
static void
InsertCacheInfo(path, atime, size)
char *path;
time_t atime;
off_t size;
{
  struct cache_info *n, *c, *t;

  n = (struct cache_info *)alloc_mem(sizeof(struct cache_info));
  n->path = alloc_string(path);
  n->atime = atime;
  n->size = size;
  n->next = NULL;

  t = NULL;
  c = clist;
  while (c)
  {
    if (c->atime > atime)
    {
      if (t != NULL) t->next = n;
      else clist = n;
      n->next = c;
      break;
    }
 
    t = c;
    c = c->next;
  }

  if (c == NULL)
  {
    if (t != NULL) t->next = n;
    else clist = n;
  }

  return;
}

/*
 * DeleteCacheInfo
 */
static void
DeleteCacheInfo(path)
char *path;
{
  struct cache_info *c, *t;

  t = NULL;
  c = clist;
  while (c)
  {
    if (strcmp(path, c->path) == 0)
    {
      if (t != NULL) t->next = c->next;
      else clist = c->next;
      break;
    }

    t = c;
    c = c->next;
  }

  if (c != NULL)
  {
    cache_size -= c->size;
    free(c->path);
    free(c);
  }

  return;
}

/*
 * InitCache
 */
void
InitCache(size, ttl, on, dir)
int size;
int ttl;
int on;
char *dir;
{
  int cache_on;
  struct stat s;
  DIR *dp;
  char *e;
  char path[MAXPATHLEN + 1];
#if defined(SYSV) || defined(SVR4)
  struct dirent *de;
#else
  struct direct *de;
#endif

  max_cache_size = ((e = getenv("CHIMERA_CACHE_SIZE")) != NULL ?
		    atoi(e):size);

  doc_ttl = (time_t)(((e = getenv("CHIMERA_DOC_TTL")) != NULL ?
                      atoi(e):ttl));

  cache_on = (getenv("CHIMERA_CACHE_OFF") == NULL && on ? 1:0);

  cache_dir = FixFilename(((e = getenv("CHIMERA_CACHE")) != NULL ?
			   e:dir));
  if (cache_dir == NULL) cache_dir = CACHE_DIR;

  if (cache_on)
  {
    dp = opendir(cache_dir);
    if (dp == NULL)
    {
      return;
    }
    
    while ((de = readdir(dp)) != NULL)
    {
      sprintf (path, "%s/%s", cache_dir, de->d_name);
      if (strncmp("ccf", de->d_name, 3) == 0 &&
	  stat(path, &s) == 0)
      {
	InsertCacheInfo(path, s.st_atime, s.st_size);
	cache_size += s.st_size;
      }
    }
    
    closedir(dp);
  }

  return;
}

/*
 * wrapstring
 */
static int
wrapstring(s)
char *s;
{
  char *cp;
  int n = 0;
  int wrap;

  for (cp = s; *cp; cp++)
  {
    n ^= *cp;
    wrap = (n >> 24) & 0xff;
    n <<= 8;
    n |= wrap;
  }

  return(n);
}

/*
 * urlToCacheName
 *
 * Jim Rees <Jim.Rees@umich.edu>
 * completely thrashed by John
 */
static char *
urlToCacheName(up)
URLParts *up;
{
  static char name[MAXPATHLEN + 1];
  int n;

  n = wrapstring(up->method);
  n += wrapstring(up->hostname);
  n += wrapstring(up->filename);
  n += wrapstring(up->ext);
  n += up->port;

  sprintf(name, "%s/ccf%x", cache_dir, n);

  return(name);
}

/*
 * IsCached
 *
 * Returns 1 if the specified URL is in the cache, otherwise returns 0.
 */
int
IsCached(url, purl)
char *url;
char *purl;
{
  struct cache_info *c;
  char *filename;
  URLParts *up;

  up = MakeURLParts(url, purl);
  if (up == NULL)
  {
    return(0);
  }

  filename = urlToCacheName(up);
  DestroyURLParts(up);

  c = clist;
  while (c)
  {
    if (strcmp(filename, c->path) == 0)
    {
      return(1);
    }
    c = c->next;
  }

  return(0);
}

/*
 * ReadCache
 */
Document *
ReadCache(up, cacheon)
URLParts *up;
int cacheon;
{
  FILE *fp;
  char buffer[BUFSIZ];
  Document *d = NULL;
  struct stat s;
  char *filename;
  int lenc;

  if (!cacheon) return(NULL);

  filename = urlToCacheName(up);

  d = CreateDocument();
  
  fp = fopen(filename, "r");
  if (fp == NULL)
  {
    return(NULL);
  }
  
  if (fstat(fileno(fp), &s) == -1)
  {
    fclose(fp);
    
    return(NULL);
  }
  
  /*
   * If the document is too old then don't use it.  Returning NULL forces
   * a reload.
   */
  if (doc_ttl && ((time((time_t *)0) - s.st_mtime) > doc_ttl))
  {
    fclose(fp);
    
    DeleteCacheInfo(filename);
    
    return(NULL);
  }
  
  fgets(buffer, sizeof(buffer), fp);
  buffer[strlen(buffer) - 1] = '\0';
  d->up = MakeURLParts(buffer, NULL);
  
  fgets(buffer, sizeof(buffer), fp);
  if (buffer[0] != '\0' && buffer[0] != '\n')
  {
    d->content = alloc_string(buffer);
    d->content[strlen(buffer) - 1] = '\0';
  }
  else
  {
    d->content = NULL;
  }
  
  fgets(buffer, sizeof(buffer), fp);
  lenc = sscanf(buffer, "%d %d", &(d->len), &(d->plen));
  
  d->text = alloc_mem(d->len + 1);
  if (fread(d->text, 1, d->len, fp) < d->len)
  {
    DestroyDocument(d);
    fclose(fp);
    unlink(filename); /* file must be messed up so get rid of it */
    return(NULL);
  }
  d->text[d->len] = '\0'; /* for text files but won't hurt binary files
			     because d->len is still the same. */
  
  if (lenc > 1)
  {
    d->ptext = alloc_mem(d->plen + 1);
    if (fread(d->ptext, 1, d->plen, fp) < d->plen)
    {
      free(d->ptext);
      d->ptext = NULL;
      d->plen = 0;
    }
    d->ptext[d->plen] = '\0';
  }
  
  fclose(fp);
  
  return(d);
}

/*
 * FlushCache
 */
static int
FlushCache()
{
  if (clist != NULL)
  {
    unlink(clist->path);
    
    DeleteCacheInfo(clist->path);

    return(1);
  }

  return(0);
}

/*
 * WriteCache
 */
void
WriteCache(d, cacheon)
Document *d;
int cacheon;
{
  FILE *fp;
  char *u;
  char *filename;
  int size = 0;

  if (d->up == NULL || d->text == NULL)
  {
    return;
  }

  filename = urlToCacheName(d->up);

  if (cacheon && d->cache)
  {
    size = d->len + (d->ptext != NULL ? d->plen:0);
    
    while (max_cache_size && ((size + cache_size) > max_cache_size))
    {
      if (!FlushCache()) return;
    }

    fp = fopen(filename, "w");
    if (fp == NULL)
    {
      return;
    }
    
    u = MakeURL(d->up, 0);
    fprintf (fp, "%s\n", u);
    free(u);
    fprintf (fp, "%s\n", (d->content != NULL ? d->content:""));
    if (d->ptext == NULL) fprintf (fp, "%d\n", d->len);
    else fprintf (fp, "%d %d\n", d->len, d->plen);
    
    fwrite(d->text, 1, d->len, fp);
    if (d->ptext != NULL) fwrite(d->ptext, 1, d->plen, fp);
    
    fclose(fp);

    cache_size += size;
  }

  InsertCacheInfo(filename, time((time_t *)0), size);
  
  return;
}

/*
 * CleanCache
 */
void
CleanCache()
{
  struct cache_info *c, *t;

  c = clist;
  while (c)
  {
    t = c;
    c = c->next;
    if (t->path)
    {
      unlink(t->path);
      free(t->path);
    }
    free(t);
  }

  return;
}
