/*
 * document.c
 *
 * Copyright (C) 1993, 1994, John Kilburg.
 *
 * You may distribute this code freely as long as it is used only
 * for academic, research, and internal business uses only, without
 * a fee.  You may distribute the binary and source code to third
 * parties provided that the copyright notice and this statement
 * appears on all copies and that no charge is associated with such
 * copies.
 *
 * If you make and distribute modified versions of the source or
 * binaries then you must inform me (John Kilburg) of the distribution.
 * You must also give me credit for the original in the source and
 * documentation but you must indicate that modifications were made.
 *
 * If you wish to make commercial use of the source or binaries you should    
 * contact me, to negotiate an appropriate license for such
 * commercial use.  Commercial use includes (1) integration of all or    
 * part of the source code into a product for sale or license by or on   
 * behalf of Licensee to third parties, or (2) distribution of the binary
 * code or source code to third parties that need it to utilize a          
 * commercial product sold or licensed by you or on your behalf.        
 * 
 * This software is provided "as is" without expressed or
 * implied warranty.  It is not intended to be used for any particular
 * purpose.  I shall not be liable for any damages suffered by users
 * of this software.  USE AT YOUR OWN RISK.
 *
 */
#include <stdio.h>
#include <ctype.h>

#ifndef NOSTDHDRS
#include <stdlib.h>
#endif

#include "url.h"
#include "access.h"
#include "document.h"
#include "local.h"
#include "util.h"
#include "conf.h"

/*
 * CreateDocument
 *
 * Allocates the structure used for documents.  Could do other
 * init stuff, too.  This function should always be called to create
 * a new document.  It provides initialization for the document which
 * is good if an old part of the program doesn't know about new
 * document attributes.
 */
Document *
CreateDocument()
{
  Document *d;

  d = (Document *)alloc_mem(sizeof(Document));
  d->ptext = NULL;
  d->text = NULL;
  d->type = DocUnknown;
  d->content = NULL;
  d->cache = 1;
  d->up = NULL;
  d->encoding = NULL;
  d->tencoding = NULL;

  return(d);
}

/*
 * TransferDocument
 */
void
TransferDocument(dest, src)
Document *dest, *src;
{
  if (dest->text) free(dest->text);
  if (dest->ptext) free(dest->ptext);
  if (dest->content) free(dest->content);
  if (dest->up) DestroyURLParts(dest->up);
  if (dest->encoding) free(dest->encoding);
  if (dest->tencoding) free(dest->tencoding);

  dest->tencoding = src->encoding;
  dest->encoding = src->encoding;
  dest->up = src->up;
  dest->text = src->text;
  dest->ptext = src->ptext;
  dest->plen = src->plen;
  dest->ptype = src->ptype;
  dest->len = src->len;
  dest->type = src->type;
  dest->content = src->content;
  dest->cache = src->cache;

  free(src);

  return;
}

/*
 * DestroyDocument
 */
void
DestroyDocument(d)
Document *d;
{
  if (d->text) free(d->text);
  if (d->content) free(d->content);
  if (d->ptext) free(d->ptext);
  if (d->up) DestroyURLParts(d->up);
  if (d->encoding) free(d->encoding);
  if (d->tencoding) free(d->tencoding);

  free(d);

  return;
}

/*
 * MakeDocument
 *
 * Creates a document using supplied HTML.
 */
Document *
MakeDocument(html)
char *html;
{
  Document *d;

  d = CreateDocument();

  d->text = alloc_string(html);
  d->len = strlen(d->text);
  d->type = DocInternal;
  d->content = alloc_string("text/html");
  d->cache = 0;

  return(d);
}

/*
 * ChangeDocument
 */
void
ChangeDocument(d, text, len, type)
Document *d;
char *text;
int len;
char *type;
{
  if (d->text)
  {
    free(d->text);
    d->text = NULL;
  }
  if (d->content)
  {
    free(d->content);
    d->content = NULL;
  }
  if (d->ptext)
  {
    free(d->ptext);
    d->ptext = NULL;
  }
  if (d->encoding)
  {
    free(d->encoding);
    d->encoding = NULL;
  }
  if (d->tencoding)
  {
    free(d->tencoding);
    d->tencoding = NULL;
  }

  d->text = alloc_mem(len + 1);
  memcpy(d->text, text, len);
  d->text[len] = '\0';
  d->len = len;
  if (type != NULL) d->content = alloc_string("text/html");

  return;
}

/*
 * Tells which functions to call for each access.
 */
struct name
{
  char *name;
  void (*func)();
  char *proxy;
};

static struct name methods[] = 
{
  { "ftp", ftp, "ftp_proxy" },
  { "http", http, "http_proxy" },
  { "gopher", gopherplain, "gopher_proxy" },
  { "gopherp", gopherplus, "gopher_proxy" },
  { "file", file, NULL },
  { "telnet", telnet, NULL },
  { "tn3270", telnet, NULL },
  { "wais", NULL, "wais_proxy" },
  { NULL, NULL, NULL },
};

/*
 * LoadDocument
 */
Document *
LoadDocument(up, reload)
URLParts *up;
int reload;
{
  int i, len;
  Document *d;
  URLParts *pup = NULL;
  char *shorturl;
  char *pproxy = NULL;
  char *no_proxy;
  char *cp, *cpe;
  int dont_proxy = 0;

  d = CreateDocument();
  d->up = DupURLParts(up);

  /*
   * Now we have the URL business figured out so pick the function which
   * handles the transfer mechanism indicated.  Returns results in d.
   */
  for (i = 0; methods[i].name != NULL; i++)
  {
    if (strcmp(up->method, methods[i].name) == 0)
    {
      if (methods[i].proxy != NULL)
      {
	no_proxy = getenv("no_proxy");
	if (no_proxy != NULL)
	{
	  for (cp = no_proxy, cpe = no_proxy; ; cpe++)
	  {
	    if (*cpe == '\0' || *cpe == ',')
	    {
	      len = strlen(up->hostname);
	      if (len >= (cpe - cp) &&
		  mystrncmp(up->hostname + len - (cpe - cp),
			    cp, cpe - cp) == 0)
	      {
		dont_proxy = 1;
		break;
	      }
	      else if (*cpe == '\0')
	      {
		break;
	      }

	      cpe++;
	      cp = cpe;
	    }
	  }
	}

	pproxy = getenv(methods[i].proxy);
	if (!dont_proxy && pproxy != NULL)
	{
	  pup = ParseURL(pproxy);
	  if (pup == NULL || NullString(pup->hostname))
	  {
	    DestroyDocument(d);
	    DestroyURLParts(pup);
	    return(NULL);
	  }
	  
	  shorturl = MakeURL(up, 0);
	  
	  http_proxy(d, pup->hostname, pup->port, shorturl, up->ext, reload);
	  
	  free(shorturl);
	  DestroyURLParts(pup);
	}
      }

      if (methods[i].func != NULL && (pproxy == NULL || dont_proxy))
      {
	(methods[i].func)(d, up->hostname, up->port, up->filename, up->ext);
      }
      break;
    }
  }

  if (d == NULL || d->text == NULL)
  {
    if (d != NULL) DestroyDocument(d); /* easier to get rid of it */

    d = MakeDocument("<h1>Error</h1>Couldn't load document");
  }
  else
  {
    if (d->up != NULL && d->up->ext != NULL && d->up->ext[0] == '?')
    {
      d->cache = 0;
    }
  }

  return(d);
}



