/*
 * url.c
 *
 * Copyright (C) 1993, 1994, John Kilburg.
 *
 * You may distribute this code freely as long as it is used only
 * for academic, research, and internal business uses only, without
 * a fee.  You may distribute the binary and source code to third
 * parties provided that the copyright notice and this statement
 * appears on all copies and that no charge is associated with such
 * copies.
 *
 * If you make and distribute modified versions of the source or
 * binaries then you must inform me (John Kilburg) of the distribution.
 * You must also give me credit for the original in the source and
 * documentation but you must indicate that modifications were made.
 *
 * If you wish to make commercial use of the source or binaries you should    
 * contact me, to negotiate an appropriate license for such
 * commercial use.  Commercial use includes (1) integration of all or    
 * part of the source code into a product for sale or license by or on   
 * behalf of Licensee to third parties, or (2) distribution of the binary
 * code or source code to third parties that need it to utilize a          
 * commercial product sold or licensed by you or on your behalf.        
 * 
 * This software is provided "as is" without expressed or
 * implied warranty.  It is not intended to be used for any particular
 * purpose.  I shall not be liable for any damages suffered by users
 * of this software.  USE AT YOUR OWN RISK.
 *
 */
#include <stdio.h>
#include <ctype.h>

#ifndef NOSTDHDRS
#include <stdlib.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <fcntl.h>

#include "url.h"
#include "util.h"
#include "conf.h"

/*
 * StripSearchURL
 *
 * Takes the ?mumble off the end.
 */
void
StripSearchURL(url)
char *url;
{
  char *cp;

  for (cp = url; *cp; cp++)
  {
    if (*cp == '?')
    {
      *cp = '\0';
      break;
    }
  }

  return;
}

/*
 * FixURL
 *
 * Fixes URLs for searches.  May do more fixing later.
 *
 * Thanks Jim.Rees@umich.edu
 */
char *
FixURL(url)
char *url;
{
  char *s, *cp;

  s = alloc_string(url);
  for (cp = s; *cp; cp++)
  {
    if (*cp == ' ')
    {
      *cp = '+';
    }
  }
  return(s);
}

/*
 * EscapeURL
 *
 * Puts escape codes in URLs.  NOT complete.
 */
char *
EscapeURL(url)
char *url;
{
  char *cp;
  char *n, *s;
  static char *hex = "0123456789ABCDEF"; /* back to my C64 days... */

  /*
   * use a bit of memory so i don't have to mess around here
   */
  s = n = alloc_string_mem(strlen(url) * 3 + 1);
  if (s == NULL)
  {
    return(NULL);
  }

  for (cp = url; *cp; cp++, n++)
  {
    if (*cp <= ' ' || *cp == '%' || *cp == '&' || *cp == '=')
    {
      *n = '%';
      n++;
      *n = hex[*cp / 16];
      n++;
      *n = hex[*cp % 16];
    }
    else
    {
      *n = *cp;
    }
  }

  *n = '\0';

  return(s);
}

/*
 * UnescapeURL
 *
 * Converts the escape codes (%xx) into actual characters.  NOT complete.
 * Could do everthing in place I guess.
 */
char *
UnescapeURL(url)
char *url;
{
  char *cp, *n, *s;
  char hex[3];

  s = n = alloc_string_mem(strlen(url) + 1);
  if (n == NULL)
  {
    return(NULL);
  }

  for (cp = url; *cp; cp++, n++)
  {
    if (*cp == '%')
    {
      cp++;
      if (*cp == '%')
      {
	*n = *cp;
      }
      else
      {
	hex[0] = *cp;
	cp++;
	hex[1] = *cp;
	hex[2] = '\0';
	*n = (char)strtol(hex, NULL, 16);
      }
    }
    else
    {
      *n = *cp;
    }
  }

  *n = '\0';

  return(s);
}

/*
 * MakeURL
 */
char *
MakeURL(up, addext)
URLParts *up;
int addext;
{
  int len;
  char *u;
  char *ext;
  char *delim;
  char *filename;
  char *hostname;
  char *method;

  if (addext && !NullString(up->ext)) ext = up->ext;
  else ext = "";

  if (NullString(up->method)) method = "file";
  else method = up->method;
  
  if (NullString(up->hostname))
  {
    delim = "";
    hostname = "";
  }
  else
  {
    delim = "//";
    hostname = up->hostname;
  }

  if (NullString(up->filename)) filename = "/";
  else filename = up->filename;

  len = strlen(method) + strlen(hostname) + strlen(filename) +
        strlen(ext) + strlen(delim) + 11;
  u = alloc_string_mem(len);
  if (up->port == -1)
  {
    sprintf (u, "%s:%s%s%s%s", method, delim, hostname,
           filename, ext);
  }
  else
  {
    sprintf (u, "%s:%s%s:%d%s%s", method, delim, hostname, up->port,
           filename, ext);
  }

  return(u);
}

/*
 * MakeURLParts
 */
URLParts *
MakeURLParts(url, context)
char *url;
char *context;
{
  URLParts *up1, *up2;
  URLParts *r;

  up1 = ParseURL(url);
  if (up1 == NULL)
  {
    return(NULL);
  }

  r = (URLParts *)alloc_mem(sizeof(URLParts));
  r->method = alloc_string("");
  r->hostname = alloc_string("");
  r->port = -1;
  r->filename = alloc_string("");
  r->ext = alloc_string("");

  /*
   * The method and hostname has been supplied or there is no context.
   */
  if ((!NullString(up1->method) && !NullString(up1->hostname)) ||
      context == NULL)
  {
    if (NullString(up1->method))
    {
      r->method = realloc_string(r->method, "file:");
      r->hostname = realloc_string(up1->hostname);
      r->port = up1->port;
      r->filename = realloc_string(up1->filename);
      r->ext = realloc_string(up1->ext);
      DestroyURLParts(up1);
      return(r);
    }
    else
    {
      DestroyURLParts(r);
      return(up1);
    }
  }

  up2 = ParseURL(context);
  if (up2 == NULL)
  {
    DestroyURLParts(r);
    return(up1);
  }
  
  /*
   * Eliminate the cases where the context can't be used.
   */
  if (!NullString(up1->method) && !NullString(up2->method))
  {
    if (strcmp(up1->method, up2->method) != 0)
    {
      DestroyURLParts(r);
      DestroyURLParts(up2);
      return(up1);
    }
  }
  
  /*
   * Use the url and the supplied context to construct a new url.
   * Jim Rees to the rescue!
   */
  if (!NullString(up1->method) || !NullString(up2->method))
  {
    r->method = realloc_string(r->method, !NullString(up1->method) ?
                             up1->method : up2->method);

    if (!NullString(up1->hostname) || !NullString(up2->hostname))
    {
      r->hostname = realloc_string(r->hostname, !NullString(up1->hostname) ?
                                 up1->hostname : up2->hostname);

      r->port = (up1->port != -1 ? up1->port:up2->port);
    }
  }
  else
  {
    r->method = realloc_string(r->method, "file:");
  }

  if (up1->filename[0] == '/' || up1->filename[0] == '~')
  {
    r->filename = realloc_string_mem(r->filename, strlen(up1->filename) + 2);
    if (up1->filename[0] != '/')
    {
      r->filename[0] = '/';
      strcpy(r->filename + 1, up1->filename);
    }
    else
    {
      strcpy(r->filename, up1->filename);
    }
  }
  else
  {
    /*
     * This includes an attempt to deal with '..' and '.'
     */
    int i;
    int len;
    int start;
    int backcount = 0;

    backcount = 0;
    start = 0;
    len = strlen(up1->filename);
    for (i = 0; i < len; i++)
    {
      if (strncmp(up1->filename + i, "../", 3) == 0)
      {
	start = i + 3;
	i += 2;
	backcount++;
      }
      else if (strncmp(up1->filename + i, "./", 2) == 0)
      {
	start = i + 2;
	i += 1;
      }
    }

    for (i = strlen(up2->filename) - 1; i >= 0; i--)
    {
      if (up2->filename[i] == '/')
      {
	if (backcount-- == 0)
	{
	  up2->filename[i + 1] = '\0';
	  break;
	}
      }
    }

    if (i == -1)
    {
      r->filename = realloc_string(r->filename, up1->filename + start);
    }
    else
    {
      r->filename = realloc_string_mem(r->filename, strlen(up2->filename) +
                                     strlen(up1->filename + start) + 1);
      strcpy(r->filename, up2->filename);
      strcat(r->filename, up1->filename + start);
    }
  }

  r->ext = realloc_string(r->ext, up1->ext);

  DestroyURLParts(up1);
  DestroyURLParts(up2);

  return(r);
}

URLParts *
DupURLParts(up)
URLParts *up;
{
  URLParts *dp;

  dp = (URLParts *)alloc_mem(sizeof(URLParts));
  dp->method = alloc_string(up->method);
  dp->hostname = alloc_string(up->hostname);
  dp->filename = alloc_string(up->filename);
  dp->ext = alloc_string(up->ext);
  dp->port = up->port;

  return(dp);
}

URLParts *
ParseURL(url)
char *url;
{
  URLParts *up;
  char *cp, *scp;
  char *of;
  int state = 0;

  up = (URLParts *)alloc_mem(sizeof(URLParts));

  up->method = NULL;
  up->hostname = NULL;
  up->port = 0;
  up->filename = NULL;
  up->ext = NULL;

  for (cp = url; isspace(*cp); cp++)
      ;

  for (scp = cp; *cp; cp++)
  {
    switch (state)
    {
      case 0: /* method */
        if (*cp == ':')
	{
	  up->method = alloc_string_mem(cp - scp + 1);
	  strncpy(up->method, scp, cp - scp);
	  up->method[cp - scp] = '\0';

	  if (*(cp + 1) == '/' && *(cp + 2) == '/')
	  {
	    state = 1;
	    cp += 2;
	  }
	  else
	  {
	    state = 3;
	  }
	  scp = cp + 1;
 	}
	break;

      case 1: /* hostname */
	if (*cp == ':' || *cp == '/')
	{
	  up->hostname = alloc_string_mem(cp - scp + 1);
	  strncpy(up->hostname, scp, cp - scp);
	  up->hostname[cp - scp] = '\0';

	  if (*cp == ':')
	  {
	    state = 2;
	    scp = cp + 1;
	  }
	  else
	  {
	    state = 3;
	    scp = cp;
	  }
	}
	break;

      case 2: /* port */
	if (!isdigit(*cp))
	{
	  state = 3;
	  scp = cp;
	}
	else
	{
	  up->port = up->port * 10 + (*cp - '0');
	}
	break;

      case 3: /* filename */
	break;
    }
  }

  /*
   * To make sure the last part gets taken care of when the
   * NULL is seen.
   */
  if (scp == url)
  {
    state = 3;
  }
  switch (state)
  {
    case 0: /* method */
      up->method = alloc_string_mem(cp - scp + 1);
      strncpy(up->method, scp, cp - scp);
      up->method[cp - scp] = '\0';
      break;
      
    case 1: /* hostname */
      up->hostname = alloc_string_mem(cp - scp + 1);
      strncpy(up->hostname, scp, cp - scp);
      up->hostname[cp - scp] = '\0';
      break;

    case 3: /* filename */
      up->filename = alloc_string_mem(cp - scp + 1);
      strncpy(up->filename, scp, cp - scp);
      up->filename[cp - scp] = '\0';

      of = NULL;
      for (cp = up->filename; *cp; cp++)
      {
	if (*cp == '#' || *cp == '?')
	{
	  of = up->filename;
	  up->filename = alloc_string_mem(cp - of + 1);
	  strncpy(up->filename, of, cp - of);
	  up->filename[cp - of] = '\0';
	  break;
	}
      }

      if (of != NULL)
      {
	for (scp = of; *cp; cp++, scp++)
	{
	  *scp = *cp;
	}
	*scp = '\0';

	up->ext = of;
      }

      break;
  }

  if (up->method == NULL) up->method = alloc_string("");
  if (up->hostname == NULL) up->hostname = alloc_string("");
  if (up->filename == NULL) up->filename = alloc_string("/");
  if (up->ext == NULL) up->ext = alloc_string("");
  if (up->port == 0) up->port = -1;

  return(up);
}

void
DestroyURLParts(up)
URLParts *up;
{
  if (up->method) free(up->method);
  if (up->hostname) free(up->hostname);
  if (up->filename) free(up->filename);
  if (up->ext) free(up->ext);
  free(up);

  return;
}
